(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (factory((global.qmidi = {})));
}(this, (function (exports) { 'use strict';

    var ErrorReporter = (function () {
        function ErrorReporter() {
            this.msgs = [];
        }
        ErrorReporter.prototype.reset = function () {
            this.msgs.length = 0;
        };
        ErrorReporter.prototype.complationError = function (msg, range) {
            this.msgs.push({ msg: msg, range: range });
        };
        ErrorReporter.prototype.forEach = function (cb) {
            for (var _i = 0, _a = this.msgs; _i < _a.length; _i++) {
                var msg = _a[_i];
                cb(msg);
            }
        };
        return ErrorReporter;
    }());

    var toneName = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'];
    var toneNum = ['1', '1#', '2', '2#', '3', '4', '4#', '5', '5#', '6', '6#', '7'];
    var numToNote = [0, 2, 4, 5, 7, 9, 11];
    var majorKeyToSignature = [
        0,
        7,
        2,
        -3,
        4,
        -1,
        6,
        1,
        -4,
        3,
        -2,
        5,
    ];
    var Note = (function () {
        function Note(note, duration) {
            this.note = note;
            this.duration = duration;
            this.velocity = 0;
        }
        Note.prototype.toString = function (useNum) {
            if (useNum === void 0) { useNum = false; }
            var n = useNum ? toneNum : toneName;
            return "< " + n[Note.getTone(this.note)] + " " + Note.getOctave(this.note) + ", " + this.duration + " >";
        };
        Note.prototype.normalize = function () {
            if (this.note < 0) {
                this.note = 0;
            }
            if (this.note > 132) {
                this.note = 131;
            }
            this.duration = Math.round(this.duration) | 0;
            if (this.duration === 0) {
                this.duration = 1;
            }
            return this;
        };
        Note.prototype.shiftOctave = function (n) {
            if (!this.isRest())
                this.note += n * 12;
        };
        Note.prototype.shift = function (n) {
            if (!this.isRest())
                this.note += n;
        };
        Note.prototype.isRest = function () {
            return this.note === Note.REST;
        };
        Note.getTone = function (note) { return note % 12; };
        Note.getOctave = function (note) { return (note / 12 | 0) - 1; };
        Note.numberToNote = function (n, octave) { return n === 0 ? 132 : numToNote[n - 1] + 12 * (octave + 1); };
        Note.shiftToKeySignature = function (shift, minor) {
            if (minor === void 0) { minor = false; }
            if (minor)
                shift -= 3;
            while (shift < 0)
                shift += 12;
            while (shift >= 12)
                shift -= 12;
            return majorKeyToSignature[shift];
        };
        Note.REST = 132;
        Note.NOTE_COUNT = 132;
        Note.DEFLEN = 96;
        return Note;
    }());
    var MidiEventType;
    (function (MidiEventType) {
        MidiEventType[MidiEventType["NOTEON"] = 0] = "NOTEON";
        MidiEventType[MidiEventType["NOTEOFF"] = 1] = "NOTEOFF";
        MidiEventType[MidiEventType["TEMPO_CHANGE"] = 2] = "TEMPO_CHANGE";
        MidiEventType[MidiEventType["KEY_SIGNATURE_CHANGE"] = 3] = "KEY_SIGNATURE_CHANGE";
        MidiEventType[MidiEventType["TIME_SIGNATURE_CHANGE"] = 4] = "TIME_SIGNATURE_CHANGE";
    })(MidiEventType || (MidiEventType = {}));
    function createNoteOnEvent(note, delta, channel, velocity) {
        return { type: MidiEventType.NOTEON, delta: delta, note: note, channel: channel, velocity: velocity };
    }
    function createNoteOffEvent(note, delta, channel, velocity) {
        return { type: MidiEventType.NOTEOFF, delta: delta, note: note, channel: channel, velocity: velocity };
    }
    function createTempoChangeEvent(tempo, delta) {
        return { type: MidiEventType.TEMPO_CHANGE, tempo: tempo, delta: delta };
    }
    function createKeySignatureChangeEvent(shift, minor, delta) {
        return { type: MidiEventType.KEY_SIGNATURE_CHANGE, shift: shift, minor: minor, delta: delta };
    }
    function createTimeSignatureChangeEvent(sig, delta) {
        return { type: MidiEventType.TIME_SIGNATURE_CHANGE, delta: delta, sig: sig };
    }
    function eventToString(e, useNum) {
        if (useNum === void 0) { useNum = false; }
        var n = useNum ? toneNum : toneName;
        switch (e.type) {
            case MidiEventType.NOTEON:
                return "NoteOn(channel = " + e.channel + ", delta = " + e.delta + ", note = " + n[Note.getTone(e.note)] + "-" + Note.getOctave(e.note) + ", velocity = " + e.velocity + ")";
            case MidiEventType.NOTEOFF:
                return "NoteOff(channel = " + e.channel + ", delta = " + e.delta + ", note = " + n[Note.getTone(e.note)] + "-" + Note.getOctave(e.note) + ", velocity = " + e.velocity + ")";
            case MidiEventType.TEMPO_CHANGE:
                return "TempoChange(delta = " + e.delta + ", tempo = " + e.tempo + ")";
            case MidiEventType.KEY_SIGNATURE_CHANGE:
                return "KeySignatureChange(delta = " + e.delta + ", shift = " + e.shift + ", minor = " + e.minor + ")";
            case MidiEventType.TIME_SIGNATURE_CHANGE:
                return "TimeSignatureChange(delta = " + e.delta + ", sig = " + e.sig.numerator + " / " + e.sig.denominator + ")";
            default:
                throw new Error('unreachable');
        }
    }
    var MidiFile = (function () {
        function MidiFile() {
            this.keysig = 0;
            this.timesig = { numerator: 4, denominator: 4 };
            this.metronome = { clocks: 24, n32: 8 };
            this.minor = false;
            this.division = Note.DEFLEN;
            this.startTempo = 500000;
            this.tracks = [];
            this.format = 1;
        }
        MidiFile.prototype.dump = function (useNum) {
            if (useNum === void 0) { useNum = false; }
            var ret = ["MidiFile(keysig = " + this.keysig + ", division = " + this.division + ", tempo = " + this.startTempo + "){"];
            for (var _i = 0, _a = this.tracks; _i < _a.length; _i++) {
                var track = _a[_i];
                ret.push("    Track(name = \"" + track.name + "\", instrument = " + track.instrument + "){");
                for (var _b = 0, _c = track.events; _b < _c.length; _b++) {
                    var event_1 = _c[_b];
                    ret.push('        ' + eventToString(event_1, useNum));
                }
                ret.push('    }');
            }
            ret.push('}');
            return ret;
        };
        MidiFile.prototype.isEmpty = function () {
            return this.tracks.length === 0;
        };
        return MidiFile;
    }());

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation. All rights reserved.
    Licensed under the Apache License, Version 2.0 (the "License"); you may not use
    this file except in compliance with the License. You may obtain a copy of the
    License at http://www.apache.org/licenses/LICENSE-2.0

    THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
    WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
    MERCHANTABLITY OR NON-INFRINGEMENT.

    See the Apache Version 2.0 License for specific language governing permissions
    and limitations under the License.
    ***************************************************************************** */
    /* global Reflect, Promise */

    var extendStatics = function(d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };

    function __extends(d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    }

    var Position = (function () {
        function Position(line, column) {
            this.line = line;
            this.column = column;
        }
        Position.prototype.reset = function () {
            this.line = this.column = 1;
        };
        Position.prototype.forward = function () {
            this.column++;
        };
        Position.prototype.newline = function () {
            this.line++;
            this.column = 1;
        };
        Position.prototype.clone = function () {
            return new Position(this.line, this.column);
        };
        return Position;
    }());
    var Range = (function () {
        function Range(start, end) {
            this.start = start;
            this.end = end;
        }
        Range.prototype.reset = function () {
            this.start.reset();
            this.end.reset();
        };
        Range.here = function (h) {
            return new Range(h, new Position(h.line, h.column + 1));
        };
        Range.between = function (p1, p2) {
            return new Range(p1.start, p2.end);
        };
        return Range;
    }());
    var TokenType;
    (function (TokenType) {
        TokenType[TokenType["EOF"] = 1] = "EOF";
        TokenType[TokenType["SPACE"] = 2] = "SPACE";
        TokenType[TokenType["MACRO"] = 3] = "MACRO";
        TokenType[TokenType["MACRO_PARAM"] = 4] = "MACRO_PARAM";
        TokenType[TokenType["BGROUP"] = 5] = "BGROUP";
        TokenType[TokenType["EGROUP"] = 6] = "EGROUP";
        TokenType[TokenType["OTHER"] = 7] = "OTHER";
    })(TokenType || (TokenType = {}));
    var Token = (function (_super) {
        __extends(Token, _super);
        function Token(type, text, start, end, hasWhiteSpace, val) {
            if (val === void 0) { val = null; }
            var _this = _super.call(this, start, end) || this;
            _this.type = type;
            _this.text = text;
            _this.hasWhiteSpace = hasWhiteSpace;
            _this.val = val;
            return _this;
        }
        Token.prototype.getText = function () {
            var t = this.text === null ? '' : this.text;
            return this.hasWhiteSpace ? ' ' + t : t;
        };
        Token.prototype.isEOF = function () {
            return this.type === TokenType.EOF;
        };
        return Token;
    }(Range));
    function repeat(s, t) {
        var r = '';
        while (t-- > 0)
            r += s;
        return r;
    }
    function markLines(lines, range, marker, space) {
        if (marker === void 0) { marker = '^'; }
        if (space === void 0) { space = ' '; }
        var ret = [];
        if (range.start.line === range.end.line) {
            ret.push(lines.getLine(range.start.line));
            var s = repeat(space, range.start.column - 1) + repeat(marker, range.end.column - range.start.column);
            ret.push(s);
        }
        else {
            var l = lines.getLine(range.start.line);
            ret.push(l);
            ret.push(repeat(space, range.start.column - 1) + repeat(marker, l.length - range.start.column));
            for (var i = range.start.line + 1; i < range.end.line - 1; i++) {
                ret.push(l = lines.getLine(i));
                ret.push(repeat(marker, l.length));
            }
            l = lines.getLine(range.end.line);
            ret.push(l);
            ret.push(repeat(marker, range.end.column));
        }
        return ret;
    }

    var TeXMacro = (function () {
        function TeXMacro(nameToken) {
            this.nameToken = nameToken;
            this.fmt = [];
            this.content = null;
            this.isMeta = false;
            this.argCount = 0;
            this.name = nameToken.text;
        }
        TeXMacro.prototype.run = function (e, macroToken) {
            var ret = [], param = new Array(this.argCount);
            var t;
            for (var i = 0, _a = this.fmt; i < _a.length; i++) {
                var f = _a[i];
                if (f.type === TokenType.MACRO_PARAM) {
                    var selectedParam = void 0;
                    if (i === _a.length - 1 || _a[i + 1].type === TokenType.MACRO_PARAM) {
                        selectedParam = param[f.val] = e.readPossibleGroup(e.nextToken(false), false);
                    }
                    else {
                        var next = _a[i + 1];
                        selectedParam = param[f.val] = [];
                        t = e.peekToken(false);
                        if (t.type !== TokenType.EOF && t.text !== next.text) {
                            while (t.type !== TokenType.EOF && t.text !== next.text) {
                                e.readPossibleGroup(e.nextToken(false), false, selectedParam);
                                t = e.peekToken(false);
                            }
                        }
                        else {
                            e.eReporter.complationError("Use of macro " + this.name + " that doesn't match its definition", t);
                            e.nextToken(false);
                            return null;
                        }
                    }
                }
                else {
                    t = e.nextToken(false);
                    if (t.type === TokenType.EOF) {
                        e.eReporter.complationError("Unexpected end of file: use of macro " + this.name + " that doesn't match its definition", t);
                        return null;
                    }
                    else if (f.text !== t.text) {
                        e.eReporter.complationError("Use of macro " + this.name + " that doesn't match its definition", t);
                        return null;
                    }
                }
            }
            for (var _i = 0, _b = this.content; _i < _b.length; _i++) {
                var tk = _b[_i];
                if (tk.type === TokenType.MACRO_PARAM) {
                    var selected = param[tk.val];
                    if (selected.length >= 1)
                        selected[0].hasWhiteSpace = tk.hasWhiteSpace;
                    for (var _c = 0, selected_1 = selected; _c < selected_1.length; _c++) {
                        var ptk = selected_1[_c];
                        ret.push(ptk);
                    }
                }
                else
                    ret.push(tk);
            }
            if (ret.length >= 1) {
                ret[0].hasWhiteSpace = ret[0].hasWhiteSpace || macroToken.hasWhiteSpace;
            }
            return new TokenArray(ret);
        };
        return TeXMacro;
    }());
    var MacroType;
    (function (MacroType) {
        MacroType[MacroType["SCOPE"] = 0] = "SCOPE";
        MacroType[MacroType["GLOBAL"] = 1] = "GLOBAL";
        MacroType[MacroType["INTERNAL"] = 2] = "INTERNAL";
    })(MacroType || (MacroType = {}));
    var MacroSet = (function () {
        function MacroSet() {
            this.macroStack = [{}];
            this.internalMacros = {};
        }
        MacroSet.prototype.getMacro = function (name) {
            for (var _a = this.macroStack, i = _a.length - 1; i >= 0; i--) {
                if (_a[i].hasOwnProperty(name))
                    return _a[i][name];
            }
            if (this.internalMacros.hasOwnProperty(name))
                return this.internalMacros[name];
            return null;
        };
        MacroSet.prototype.reset = function () {
            this.macroStack = [{}];
        };
        MacroSet.prototype.defineMacro = function (m, type) {
            if (type === MacroType.SCOPE) {
                this.macroStack[this.macroStack.length - 1][m.name] = m;
            }
            else if (type === MacroType.GLOBAL) {
                this.macroStack[0][m.name] = m;
            }
            else if (type === MacroType.INTERNAL) {
                this.internalMacros[m.name] = m;
            }
            return this;
        };
        MacroSet.prototype.define = function (name, run, type) {
            this.defineMacro({ name: name, run: run, isMeta: false }, type);
            return this;
        };
        MacroSet.prototype.defineMeta = function (name) {
            this.defineMacro({ name: name, run: null, isMeta: true }, MacroType.INTERNAL);
            return this;
        };
        MacroSet.prototype.enterScope = function () {
            this.macroStack.push({});
            return this;
        };
        MacroSet.prototype.leaveScope = function () {
            this.macroStack.pop();
            return this;
        };
        MacroSet.prototype.isGlobal = function () {
            return this.macroStack.length === 0;
        };
        MacroSet.prototype.defineInternalMacros = function () {
            var cela = this;
            function def(e, global) {
                var t = e.nextToken(false);
                if (t.type !== TokenType.MACRO) {
                    e.eReporter.complationError('Macro name expected', t);
                    return null;
                }
                var macro = new TeXMacro(t);
                t = e.nextToken(false);
                while (t.type !== TokenType.BGROUP && t.type !== TokenType.EOF) {
                    if (t.type === TokenType.MACRO_PARAM) {
                        if (t.val !== ++macro.argCount) {
                            e.eReporter.complationError('Macro parameter number must be consecutive', t);
                        }
                        macro.fmt.push(t);
                    }
                    else
                        macro.fmt.push(t);
                    t = e.nextToken(false);
                }
                if (t.type !== TokenType.BGROUP) {
                    e.eReporter.complationError('"{" expected', t);
                    return null;
                }
                macro.content = e.readPossibleGroup(t, false);
                cela.defineMacro(macro, global ? MacroType.GLOBAL : MacroType.SCOPE);
                return null;
            }
            this.define('\\def', function (e) { return def(e, false); }, MacroType.INTERNAL);
            this.define('\\gdef', function (e) { return def(e, true); }, MacroType.INTERNAL);
            return this;
        };
        return MacroSet;
    }());
    var TokenArray = (function () {
        function TokenArray(ta) {
            this.ta = ta;
            this.i = 0;
        }
        TokenArray.prototype.nextToken = function () { return this.i >= this.ta.length ? null : this.ta[this.i++]; };
        TokenArray.prototype.peekToken = function () { return this.i >= this.ta.length ? null : this.ta[this.i]; };
        return TokenArray;
    }());
    var MacroExpander = (function () {
        function MacroExpander(reporter, tSource, macroSet) {
            this.maxNestedMacro = 100;
            this._tk = null;
            this.processStack = [];
            this.eReporter = reporter;
            this.macros = macroSet === void 0 ? new MacroSet() : macroSet;
            this.processStack = tSource ? [tSource] : [];
        }
        MacroExpander.prototype.init = function (ts) {
            this.processStack.length = 0;
            this._tk = null;
            this.processStack.push(ts);
            this.macros.reset();
        };
        MacroExpander.prototype._expand = function (tk, macro) {
            if (macro === null) {
                this.eReporter.complationError("Undefined control sequence " + tk.text, tk);
                return;
            }
            var ts = macro.run(this, tk);
            if (ts !== null) {
                if (this.processStack.length >= this.maxNestedMacro) {
                    this.eReporter.complationError('Maximum nested macro expansion exceeded', tk);
                    this.processStack.length = 1;
                }
                else {
                    this.processStack.push(ts);
                }
            }
        };
        MacroExpander.prototype._pull = function () {
            var t = this.processStack[this.processStack.length - 1].nextToken();
            while (t === null) {
                this.processStack.pop();
                t = this.processStack[this.processStack.length - 1].nextToken();
            }
            return t;
        };
        MacroExpander.prototype.readPossibleGroup = function (bg, expand, array) {
            var ret;
            if (array === void 0)
                ret = [bg];
            else {
                ret = array;
                ret.push(bg);
            }
            var level = 1;
            if (bg.type === TokenType.BGROUP) {
                while (level > 0) {
                    var t = this.nextToken(expand);
                    if (t.type === TokenType.EOF) {
                        this.eReporter.complationError('missing "}"', t);
                        level = 0;
                    }
                    else {
                        ret.push(t);
                        t.type === TokenType.BGROUP && level++;
                        t.type === TokenType.EGROUP && level--;
                    }
                }
            }
            return ret;
        };
        MacroExpander.prototype._readToken = function (expand) {
            if (expand === void 0) { expand = true; }
            do {
                var ret = this._pull();
                if (expand && ret.type === TokenType.MACRO) {
                    var macro = this.macros.getMacro(ret.text);
                    if (macro === null || (macro !== null && !macro.isMeta)) {
                        this._expand(ret, macro);
                        continue;
                    }
                    else
                        return ret;
                }
                else {
                    if (ret.type === TokenType.BGROUP) {
                        this.macros.enterScope();
                    }
                    else if (ret.type === TokenType.EGROUP) {
                        if (this.macros.isGlobal()) {
                            this.eReporter.complationError('"}" without "{"', ret);
                        }
                        else
                            this.macros.leaveScope();
                    }
                    if (ret.type === TokenType.EOF && this.macros.isGlobal()) {
                        this.eReporter.complationError('missing "}"', ret);
                    }
                    return ret;
                }
            } while (true);
        };
        MacroExpander.prototype.nextToken = function (expand) {
            if (expand === void 0) { expand = true; }
            if (this._tk === null) {
                return this._readToken(expand);
            }
            else {
                var tk = this._tk;
                this._tk = null;
                return tk;
            }
        };
        MacroExpander.prototype.peekToken = function (expand) {
            if (expand === void 0) { expand = true; }
            return this._tk === null ? this._tk = this._readToken(expand) : this._tk;
        };
        return MacroExpander;
    }());

    var List = (function () {
        function List() {
            this.head = null;
            this.tail = null;
        }
        List.prototype.remove = function (node) {
            if (node.prev)
                node.prev.next = node.next;
            if (node.next)
                node.next.prev = node.prev;
            if (node === this.head)
                this.head = node.next;
            if (node === this.tail)
                this.tail = node.prev;
        };
        List.prototype.add = function (data) {
            var node = { data: data, prev: this.tail, next: null };
            if (!this.head)
                this.head = node;
            if (this.tail)
                this.tail.next = node;
            this.tail = node;
        };
        List.prototype.forEach = function (cb) {
            for (var n = this.head; n; n = n.next) {
                cb(n.data, n);
            }
        };
        return List;
    }());

    var ModifierNodeType;
    (function (ModifierNodeType) {
        ModifierNodeType[ModifierNodeType["VIA"] = 0] = "VIA";
        ModifierNodeType[ModifierNodeType["DASH"] = 1] = "DASH";
        ModifierNodeType[ModifierNodeType["UNDERLINE"] = 2] = "UNDERLINE";
        ModifierNodeType[ModifierNodeType["SHARP"] = 3] = "SHARP";
        ModifierNodeType[ModifierNodeType["FLAT"] = 4] = "FLAT";
        ModifierNodeType[ModifierNodeType["DOT"] = 5] = "DOT";
        ModifierNodeType[ModifierNodeType["POS_OCTAVE"] = 6] = "POS_OCTAVE";
        ModifierNodeType[ModifierNodeType["NEG_OCTAVE"] = 7] = "NEG_OCTAVE";
        ModifierNodeType[ModifierNodeType["SLASH"] = 8] = "SLASH";
    })(ModifierNodeType || (ModifierNodeType = {}));
    var modifierToType = {
        '_': ModifierNodeType.UNDERLINE,
        '-': ModifierNodeType.DASH,
        '#': ModifierNodeType.SHARP,
        'b': ModifierNodeType.FLAT,
        '.': ModifierNodeType.NEG_OCTAVE,
        '\'': ModifierNodeType.POS_OCTAVE,
        '*': ModifierNodeType.DOT,
        '/': ModifierNodeType.SLASH
    };
    var modifierWithVal = {
        '-': ModifierNodeType.DASH,
        '*': ModifierNodeType.DOT,
        '/': ModifierNodeType.SLASH
    };
    var modifierNodeTypeToLiteral = ['', '-', '_', '#', 'b', '*', '\'', '.', '/'];
    var EventNodeType;
    (function (EventNodeType) {
        EventNodeType[EventNodeType["NOTE"] = 0] = "NOTE";
        EventNodeType[EventNodeType["REST"] = 1] = "REST";
        EventNodeType[EventNodeType["TEMPO_CHANGE"] = 2] = "TEMPO_CHANGE";
        EventNodeType[EventNodeType["KEY_SIGNATURE_CHANGE"] = 3] = "KEY_SIGNATURE_CHANGE";
        EventNodeType[EventNodeType["TIME_SIGNATURE_CHANGE"] = 4] = "TIME_SIGNATURE_CHANGE";
    })(EventNodeType || (EventNodeType = {}));
    function createNoteNode(note, velocity, channel, pos) {
        return { type: EventNodeType.NOTE, next: null, sibling: null, parent: null, note: note, pos: pos, velocity: velocity, channel: channel, refCount: 1 };
    }
    function createModifierNode(type, pos) {
        if (pos === void 0) { pos = null; }
        return { parent: null, type: type, pos: pos };
    }
    function createModifierNodeWithVal(type, pos, val) {
        return { type: type, parent: null, pos: pos, val: val };
    }
    function createRestEventNode(pos) {
        return { type: EventNodeType.REST, next: null, sibling: null, parent: null, pos: pos, refCount: 1 };
    }
    function createKeySignatureChangeNode(pos, shift, minor) {
        return { type: EventNodeType.KEY_SIGNATURE_CHANGE, next: null, sibling: null, parent: null, pos: pos, shift: shift, minor: minor, refCount: 1 };
    }
    function createTempoChangeNode(pos, tempo) {
        return { type: EventNodeType.TEMPO_CHANGE, next: null, sibling: null, parent: null, pos: pos, tempo: tempo, refCount: 1 };
    }
    function createTimeSignatureChangeNode(pos, sig) {
        return { type: EventNodeType.TIME_SIGNATURE_CHANGE, next: null, sibling: null, parent: null, pos: pos, sig: sig, refCount: 1 };
    }
    var dummyNodeSlot = { insertNode: function (node) { } };
    var EventNodeList = (function () {
        function EventNodeList() {
            this.head = null;
            this.tail = null;
            this.sibling = null;
            this.topNode = null;
            this._freeChildren = [];
        }
        EventNodeList.prototype._copyFrom = function (list) {
            this.head = list.head;
            this.tail = list.tail;
            this.sibling = list.sibling;
            this.topNode = list.topNode;
            this._freeChildren = list._freeChildren;
        };
        EventNodeList.prototype._connectTailTo = function (node) {
            this.tail.next = node;
            for (var _i = 0, _a = this._freeChildren; _i < _a.length; _i++) {
                var n = _a[_i];
                n.next = node;
            }
            node.refCount = this._freeChildren.length + 1;
            this._freeChildren.length = 0;
        };
        EventNodeList.prototype.append = function (node) {
            if (this.head) {
                this._connectTailTo(node);
                this.tail = node;
            }
            else {
                this.sibling = this.head = this.tail = node;
                this.topNode = createModifierNode(ModifierNodeType.VIA);
            }
            var top = node.parent = this.topNode;
            var cur = null;
            return {
                insertNode: function (n) {
                    n.parent = top;
                    if (cur) {
                        cur.parent = n;
                    }
                    else {
                        node.parent = n;
                    }
                    cur = n;
                }
            };
        };
        EventNodeList.prototype.appendChild = function (list, notDiverse) {
            if (notDiverse === void 0) { notDiverse = false; }
            if (list.head) {
                if (this.head) {
                    this.sibling.sibling = list.head;
                    this.sibling = list.sibling;
                    list.topNode.parent = this.topNode;
                    notDiverse || this._freeChildren.push(list.tail);
                }
                else {
                    this._copyFrom(list);
                    this.topNode = list.topNode.parent = createModifierNode(ModifierNodeType.VIA);
                }
            }
        };
        EventNodeList.prototype.concat = function (list) {
            if (list.head) {
                if (this.head) {
                    this._connectTailTo(list.head);
                    this.tail = list.tail;
                    list.topNode.parent = this.topNode;
                }
                else {
                    this._copyFrom(list);
                    this.topNode = list.topNode.parent = createModifierNode(ModifierNodeType.VIA);
                }
                var top_1 = this.topNode;
                var cur_1 = list.topNode;
                return {
                    insertNode: function (node) {
                        node.parent = top_1;
                        cur_1.parent = node;
                        cur_1 = node;
                    }
                };
            }
            else
                return dummyNodeSlot;
        };
        return EventNodeList;
    }());
    var regDigit = /[0-9]/;
    var regNum = /^[0-9]+$/;
    var majorKeyNameToShift = {
        'Cb': -1,
        'C': 0,
        'C#': 1,
        'Db': 1,
        'D': 2,
        'D#': 3,
        'Eb': 3,
        'E': 4,
        'E#': 5,
        'F': 5,
        'F#': 6,
        'Gb': 6,
        'G': 7,
        'G#': 8,
        'Ab': -4,
        'A': -3,
        'A#': -2,
        'Bb': -2,
        'B': -1,
        'B#': 0,
    };
    function createParser(eReporter) {
        var macroExpander = new MacroExpander(eReporter);
        var ctx = [];
        macroExpander.macros
            .defineInternalMacros()
            .defineMeta('\\tempo')
            .defineMeta('\\track')
            .defineMeta('\\v')
            .defineMeta('\\instrument')
            .defineMeta('\\bpm')
            .defineMeta('\\major')
            .defineMeta('\\minor')
            .defineMeta('\\vel')
            .defineMeta('\\times')
            .defineMeta('\\div');
        return {
            parse: parse
        };
        function next() {
            return macroExpander.nextToken();
        }
        function peek() {
            return macroExpander.peekToken();
        }
        function isMacro(tk, name) {
            return tk.type === TokenType.MACRO && tk.text === '\\' + name;
        }
        function readArg() {
            var tks = macroExpander.readPossibleGroup(next(), true);
            if (tks[0].type === TokenType.EOF) {
                eReporter.complationError('argument expected', tks[0]);
                tks[0].text = '';
                return tks[0];
            }
            else if (tks.length === 1) {
                return tks[0];
            }
            else {
                var text = '';
                for (var i = 1; i < tks.length - 1; i++) {
                    text += tks[i].getText();
                }
                return new Token(TokenType.OTHER, text, tks[1].start, tks[tks.length - 2].end, tks[0].hasWhiteSpace);
            }
        }
        function readNumber() {
            var num = readArg();
            if (regNum.test(num.text)) {
                return num;
            }
            else {
                eReporter.complationError('Number expected', num);
                return null;
            }
        }
        function createTrack(name, channel, file) {
            return { voices: {}, volume: 0x64, name: name, instrument: -1, shift: file.keysig, channel: channel };
        }
        function createDefaultContext(file, channel) {
            return { velocity: 80, octave: 4, shift: file.keysig, channel: channel };
        }
        function createVoice(file, channel) {
            return { channel: channel, ctx: createDefaultContext(file, channel), events: new EventNodeList() };
        }
        function enterScope(nctx) {
            var t;
            if (nctx) {
                ctx.push(nctx);
                return nctx;
            }
            else {
                t = scopeTop();
                ctx.push(t = { velocity: t.velocity, octave: t.octave, shift: t.shift, channel: t.channel });
                return t;
            }
        }
        function leaveScope() {
            ctx.pop();
        }
        function scopeTop() {
            return ctx[ctx.length - 1];
        }
        function parse(tkSource) {
            macroExpander.init(tkSource);
            ctx.length = 0;
            var _a = parseTop(), tracks = _a.tracks, file = _a.file;
            for (var name_1 in tracks) {
                var track = convertTrack(file, tracks[name_1]);
                track.events.length > 0 && file.tracks.push(track);
            }
            return file;
        }
        function convertTrack(file, track) {
            var list = new EventNodeList();
            for (var l in track.voices) {
                list.appendChild(track.voices[l].events, true);
            }
            return {
                name: track.name,
                volume: track.volume,
                instrument: track.instrument,
                events: pollAllEvents(createOverlappedEventMerger(createNoteQueue(list, file.division)))
            };
        }
        function parseTop() {
            var ret = { tracks: {}, file: new MidiFile() };
            var tracks = ret.tracks;
            var trackCount = 0;
            parseFileOptions(ret.file);
            var tk = peek();
            if (tk.type !== TokenType.EOF) {
                var name_2;
                if (!isMacro(tk, 'track')) {
                    name_2 = 'Track 1';
                    if (!tracks.hasOwnProperty(name_2)) {
                        parseTrackContent(ret.file, tracks[name_2] = createTrack(name_2, trackCount++, ret.file));
                    }
                    else
                        parseTrackContent(ret.file, tracks[name_2]);
                    tk = peek();
                }
                while (isMacro(tk, 'track')) {
                    next();
                    name_2 = readArg().text;
                    if (!tracks.hasOwnProperty(name_2)) {
                        parseTrackContent(ret.file, tracks[name_2] = createTrack(name_2, trackCount++, ret.file));
                    }
                    else
                        parseTrackContent(ret.file, tracks[name_2]);
                    tk = peek();
                }
            }
            return ret;
        }
        function parseTempoMacro() {
            next();
            var n = readNumber();
            if (n) {
                var tempo = Number(n.text);
                if (tempo > 0xffffff) {
                    eReporter.complationError('Tempo value too large, should be less than 16777215 (0xffffff)', n);
                    return -1;
                }
                else
                    return tempo;
            }
            else {
                return -1;
            }
        }
        function parseBpmMacro() {
            next();
            var n = readArg();
            if (regNum.test(n.text)) {
                var tempo = 60000000 / Number(n.text);
                if (tempo > 0xffffff) {
                    eReporter.complationError("Tempo value too large (" + tempo + "), should be less than 16777215 (0xffffff)", n);
                    return -1;
                }
                else
                    return tempo;
            }
            else {
                eReporter.complationError('BPM number expected', n);
                return -1;
            }
        }
        function parseTimeSignatureMacro() {
            next();
            var numerator = readNumber();
            var denominator = readNumber();
            if (numerator !== null && denominator !== null) {
                return { numerator: Number(numerator.text), denominator: Number(denominator.text) };
            }
            else {
                return null;
            }
        }
        function parseMajorOrMinorMacro() {
            var minor = isMacro(next(), 'minor');
            var keyName = readArg();
            if (majorKeyNameToShift.hasOwnProperty(keyName.text)) {
                var shift = majorKeyNameToShift[keyName.text];
                minor && (shift -= 3);
                return { shift: shift, minor: minor };
            }
            else {
                eReporter.complationError("Unknwon key signature name " + keyName.text, keyName);
                return null;
            }
        }
        function parseFileOptions(file) {
            var tk = peek();
            while (true) {
                if (isMacro(tk, 'tempo')) {
                    var tempo = parseTempoMacro();
                    if (tempo !== -1) {
                        file.startTempo = tempo;
                    }
                }
                else if (isMacro(tk, 'bpm')) {
                    var tempo = parseBpmMacro();
                    if (tempo !== -1) {
                        file.startTempo = tempo;
                    }
                }
                else if (isMacro(tk, 'major') || isMacro(tk, 'minor')) {
                    var sig = parseMajorOrMinorMacro();
                    if (sig) {
                        file.keysig = sig.shift;
                        file.minor = sig.minor;
                    }
                }
                else if (isMacro(tk, 'times')) {
                    var s = parseTimeSignatureMacro();
                    if (s) {
                        file.timesig = s;
                    }
                }
                else if (isMacro(tk, 'div')) {
                    next();
                    var t = readNumber();
                    if (t) {
                        var div = Number(t.text);
                        if (div >= 0x7fff) {
                            eReporter.complationError('Division value too large, should be no more than 32767 (0x7fff)', t);
                        }
                        else {
                            file.division = div;
                        }
                    }
                }
                else
                    break;
                tk = peek();
            }
        }
        function isTrackEnd(tk) {
            return tk.type === TokenType.EOF ||
                isMacro(tk, 'track');
        }
        function parseVoice(file, track, name) {
            var v;
            if (track.voices.hasOwnProperty(name)) {
                v = track.voices[name];
            }
            else {
                v = track.voices[name] = createVoice(file, track.channel);
            }
            enterScope(v.ctx);
            v.events.concat(parseSequence());
            leaveScope();
        }
        function parseTrackContent(file, track) {
            parseTrackOptions(track);
            var tk = peek();
            if (!isTrackEnd(tk)) {
                if (!isMacro(tk, 'v')) {
                    parseVoice(file, track, '1');
                    tk = peek();
                }
                while (isMacro(tk, 'v')) {
                    next();
                    var name_3 = readArg();
                    parseVoice(file, track, name_3.text);
                    tk = peek();
                }
            }
        }
        function parseTrackOptions(track) {
            var tk = peek();
            track.instrument = 0;
            while (true) {
                if (isMacro(tk, 'instrument')) {
                    next();
                    var name_4 = readArg();
                    if (regNum.test(name_4.text))
                        track.instrument = Number(name_4.text);
                    else {
                        eReporter.complationError("Unknown instrument number " + name_4.text, name_4);
                        track.instrument = 0;
                    }
                    tk = peek();
                }
                else
                    break;
            }
        }
        function parseGroup() {
            next();
            enterScope();
            var list = parseSequence();
            leaveScope();
            var tk = peek();
            while (tk.type !== TokenType.EOF && tk.text === ',') {
                next();
                enterScope();
                list.appendChild(parseSequence());
                leaveScope();
                tk = peek();
            }
            if (next().type !== TokenType.EGROUP) {
                eReporter.complationError("'}' expected", tk);
            }
            return list;
        }
        function isSequenceEnd(tk) {
            return tk.type === TokenType.EOF ||
                tk.type === TokenType.EGROUP ||
                tk.type === TokenType.OTHER && tk.text === ',' ||
                isMacro(tk, 'v') ||
                isMacro(tk, 'track');
        }
        function parseSequence() {
            var tk = peek();
            var list = new EventNodeList();
            while (!isSequenceEnd(tk)) {
                if (tk.type === TokenType.OTHER && tk.text === '|') {
                    next();
                }
                else {
                    var slot = null;
                    if (tk.type === TokenType.BGROUP) {
                        var group = parseGroup();
                        slot = list.concat(group);
                    }
                    else if (tk.type === TokenType.MACRO) {
                        var node = parseDirective();
                        node && list.append(node);
                    }
                    else {
                        slot = list.append(parseNote(tk));
                    }
                    parseNoteModifiers(slot);
                }
                tk = peek();
            }
            return list;
        }
        function parseDirective() {
            var tk = peek();
            var top = scopeTop();
            if (isMacro(tk, 'minor') || isMacro(tk, 'major')) {
                var k = parseMajorOrMinorMacro();
                if (k !== null) {
                    top.shift = k.shift;
                    return createKeySignatureChangeNode(tk, k.shift, k.minor);
                }
            }
            else if (isMacro(tk, 'bpm')) {
                var bpm = parseBpmMacro();
                if (bpm !== -1) {
                    return createTempoChangeNode(tk, bpm);
                }
                else
                    return null;
            }
            else if (isMacro(tk, 'tempo')) {
                var tempo = parseTempoMacro();
                if (tempo !== -1)
                    return createTempoChangeNode(tk, tempo);
                else
                    return null;
            }
            else if (isMacro(tk, 'vel')) {
                next();
                var n = readNumber();
                if (n) {
                    var vel = Number(n.text);
                    if (vel > 0x7f) {
                        eReporter.complationError('Velocity value too large, it should be less than 127', n);
                    }
                    else {
                        top.velocity = vel;
                    }
                }
                return null;
            }
            else if (isMacro(tk, 'times')) {
                var s = parseTimeSignatureMacro();
                if (s) {
                    return createTimeSignatureChangeNode(tk, s);
                }
                else
                    return null;
            }
            else {
                throw new Error("Unreachable: Unimplemented directive " + tk.text);
            }
        }
        function parseNote(tk) {
            var top = scopeTop();
            if (regDigit.test(tk.text)) {
                next();
                var n = Number(tk.text);
                if (n >= 1 && n <= 7) {
                    return createNoteNode(Note.numberToNote(n, top.octave) + top.shift, top.velocity, top.channel, tk);
                }
                else if (n === 0) {
                    return createRestEventNode(tk);
                }
                else {
                    eReporter.complationError("Unknown note \"" + n + "\", valid notes are 0-7", tk);
                    return createNoteNode(-1, 0, 0, tk);
                }
            }
            else {
                next();
                eReporter.complationError('Note expected', tk);
                return createNoteNode(-1, 0, 0, tk);
            }
        }
        function parseNoteModifiers(slot) {
            var tk = peek();
            while (tk.type === TokenType.OTHER && tk.text !== ',' && modifierToType.hasOwnProperty(tk.text)) {
                var type = modifierToType[tk.text];
                if (modifierWithVal.hasOwnProperty(tk.text)) {
                    var s = tk.text;
                    var start = tk, end = tk;
                    var num = 1;
                    next();
                    tk = peek();
                    while (tk.text === s) {
                        num++;
                        end = tk;
                        next();
                        tk = peek();
                    }
                    slot.insertNode(createModifierNodeWithVal(type, Range.between(start, end), num));
                }
                else {
                    slot.insertNode(createModifierNode(modifierToType[tk.text], tk));
                    next();
                    tk = peek();
                }
            }
        }
    }
    function getModifiers(note) {
        var s = '';
        for (var n = note.parent; n; n = n.parent) {
            s += modifierNodeTypeToLiteral[n.type];
        }
        return s;
    }
    function dumpNoteList(list) {
        var s = '';
        for (var node = list.head; node; node = node.next) {
            if (node.type === EventNodeType.NOTE) {
                s += node.note + getModifiers(node) + ' ';
            }
        }
        return [s];
    }
    function createNoteQueue(list, division) {
        var queue = new List();
        pushNote(list.head, 0);
        return {
            pollNote: pollNote
        };
        function getNoteFromNode(node) {
            var ret = new Note(node.type === EventNodeType.NOTE ? node.note : Note.REST, division);
            for (var n = node.parent; n; n = n.parent) {
                switch (n.type) {
                    case ModifierNodeType.DASH:
                        ret.duration *= n.val + 1;
                        break;
                    case ModifierNodeType.SLASH:
                        ret.duration /= n.val + 1;
                        break;
                    case ModifierNodeType.UNDERLINE:
                        ret.duration >>= 1;
                        break;
                    case ModifierNodeType.POS_OCTAVE:
                        ret.shiftOctave(1);
                        break;
                    case ModifierNodeType.NEG_OCTAVE:
                        ret.shiftOctave(-1);
                        break;
                    case ModifierNodeType.SHARP:
                        ret.shift(1);
                        break;
                    case ModifierNodeType.FLAT:
                        ret.shift(-1);
                        break;
                    case ModifierNodeType.DOT:
                        var factor = 1 << n.val;
                        ret.duration = ret.duration * (2 * factor - 1) / factor;
                        break;
                }
            }
            return ret.normalize();
        }
        function pushNote(node, delta) {
            if (node && --node.refCount === 0) {
                var stack = [{ node: node, delta: delta }];
                while (stack.length > 0) {
                    var top_2 = stack.pop();
                    delta = top_2.delta;
                    for (node = top_2.node; node; node = node.sibling) {
                        if (node.type === EventNodeType.NOTE) {
                            var note = getNoteFromNode(node);
                            queue.add({ node: null, event: createNoteOnEvent(note.note, delta, node.channel, node.velocity) });
                            queue.add({ node: node, event: createNoteOffEvent(note.note, delta + note.duration, node.channel, node.velocity) });
                        }
                        else if (node.type === EventNodeType.REST) {
                            if (node.next && --node.next.refCount === 0) {
                                var note = getNoteFromNode(node);
                                stack.push({ node: node.next, delta: delta + note.duration });
                            }
                        }
                        else if (node.type === EventNodeType.KEY_SIGNATURE_CHANGE) {
                            queue.add({ node: node, event: createKeySignatureChangeEvent(node.shift, node.minor, delta) });
                        }
                        else if (node.type === EventNodeType.TEMPO_CHANGE) {
                            queue.add({ node: node, event: createTempoChangeEvent(node.tempo, delta) });
                        }
                        else if (node.type === EventNodeType.TIME_SIGNATURE_CHANGE) {
                            queue.add({ node: node, event: createTimeSignatureChangeEvent(node.sig, delta) });
                        }
                        else
                            throw new Error('unreachable');
                    }
                }
            }
        }
        function pollNote() {
            var first = queue.head;
            if (first) {
                for (var n = first.next; n; n = n.next) {
                    if (n.data.event.delta < first.data.event.delta) {
                        first = n;
                    }
                }
                var ret_1 = first.data;
                queue.remove(first);
                queue.forEach(function (n) { return n.event.delta -= ret_1.event.delta; });
                if (ret_1.node) {
                    pushNote(ret_1.node.next, 0);
                }
                return ret_1.event;
            }
            else
                return null;
        }
    }
    function createMidiEventFilter(queue, filter) {
        return {
            pollNote: function () {
                var event = queue.pollNote();
                var delta = 0;
                while (event && filter(event)) {
                    delta += event.delta;
                    event = queue.pollNote();
                }
                if (event)
                    event.delta += delta;
                return event;
            }
        };
    }
    function createOverlappedEventMerger(queue) {
        var noteRefCounts = [];
        for (var i = 0; i < Note.NOTE_COUNT; i++) {
            noteRefCounts.push(0);
        }
        return createMidiEventFilter(queue, function (event) {
            if (event.type === MidiEventType.NOTEON) {
                return noteRefCounts[event.note]++ !== 0;
            }
            else if (event.type === MidiEventType.NOTEOFF) {
                return --noteRefCounts[event.note] !== 0;
            }
            else
                return false;
        });
    }
    function pollAllEvents(queue) {
        var events = [];
        var event = queue.pollNote();
        while (event) {
            events.push(event);
            event = queue.pollNote();
        }
        return events;
    }

    var regWhiteSpace = /[ \t\r\n]/;
    var regName = /[a-zA-Z]/;
    var regDigit$1 = /[0-9]/;
    var Scanner = (function () {
        function Scanner(_source) {
            if (_source === void 0) { _source = null; }
            this._source = _source;
            this.pos = new Position(1, 1);
            this._tk = null;
            this.macroChar = '\\';
            this.macroParamChar = '#';
            this.bgroupChar = '{';
            this.egroupChar = '}';
            this.commentChar = '%';
        }
        Scanner.prototype.reset = function (s) {
            if (s === void 0) { s = null; }
            this.pos.reset();
            this._tk = null;
            if (s !== null) {
                this._source = s;
            }
        };
        Scanner.prototype._next = function () {
            var c = this._source.next();
            if (c === '\r') {
                this._source.peek() === '\n' && this._source.next();
                return '\n';
            }
            return c;
        };
        Scanner.prototype._consume = function (c) {
            if (c === '\n')
                this.pos.newline();
            else {
                this.pos.forward();
                c.charCodeAt(0) >= 0x7f && this.pos.forward();
            }
        };
        Scanner.prototype._isLetter = function (c) {
            return regName.test(c);
        };
        Scanner.prototype.nextToken = function () {
            if (this._tk === null) {
                return this._scanToken();
            }
            else {
                var tk = this._tk;
                this._tk = null;
                return tk;
            }
        };
        Scanner.prototype.peekToken = function () {
            return this._tk === null ? this._tk = this._scanToken() : this._tk;
        };
        Scanner.prototype._scanToken = function () {
            var c = this._next();
            var noWhiteSpace = false, hasWhiteSpace = false;
            do {
                if (regWhiteSpace.test(c)) {
                    hasWhiteSpace = true;
                    while (regWhiteSpace.test(c)) {
                        this._consume(c);
                        c = this._next();
                    }
                }
                else if (c === this.commentChar) {
                    this._consume(c);
                    c = this._next();
                    hasWhiteSpace = true;
                    while (c !== null && c !== '\n') {
                        this._consume(c);
                        c = this._next();
                    }
                }
                else
                    noWhiteSpace = true;
            } while (!noWhiteSpace);
            var cur = this.pos.clone();
            if (c === null) {
                return new Token(TokenType.EOF, null, cur, this.pos.clone(), hasWhiteSpace);
            }
            if (c === this.macroChar) {
                this._consume(c);
                if (this._isLetter(this._source.peek())) {
                    var name_1 = this.macroChar + this._source.next();
                    this._consume(c);
                    while (this._isLetter(c = this._source.peek()) && c !== null) {
                        this._consume(c);
                        name_1 += this._next();
                    }
                    return new Token(TokenType.MACRO, name_1, cur, this.pos.clone(), hasWhiteSpace);
                }
                else {
                    return new Token(TokenType.OTHER, c, cur, this.pos.clone(), hasWhiteSpace);
                }
            }
            else if (c === this.macroParamChar) {
                this._consume(c);
                if (regDigit$1.test(this._source.peek())) {
                    c = this._source.next();
                    this._consume(c);
                    return new Token(TokenType.MACRO_PARAM, null, cur, this.pos.clone(), hasWhiteSpace, Number(c));
                }
                else
                    return new Token(TokenType.OTHER, c, cur, this.pos.clone(), hasWhiteSpace);
            }
            else {
                this._consume(c);
                var type = void 0;
                var ch = c.charAt(0);
                if (c === this.bgroupChar) {
                    type = TokenType.BGROUP;
                }
                else if (c === this.egroupChar) {
                    type = TokenType.EGROUP;
                }
                else {
                    type = TokenType.OTHER;
                }
                return new Token(type, c, cur, this.pos.clone(), hasWhiteSpace);
            }
        };
        return Scanner;
    }());

    var Chunk = (function () {
        function Chunk() {
            this.data = [];
        }
        Chunk.prototype.writeBytes = function (bytes) {
            for (var _i = 0, bytes_1 = bytes; _i < bytes_1.length; _i++) {
                var b = bytes_1[_i];
                this.data.push(b);
            }
            return this;
        };
        Chunk.prototype.writeWord = function (n) {
            n &= 0xffff;
            this.data.push(n >> 8);
            this.data.push(n & 0xff);
            return this;
        };
        Chunk.prototype.writeString = function (s, limit) {
            if (limit === void 0) { limit = -1; }
            for (var i = 0; i < s.length && (limit === -1 || i < limit); i++) {
                this.data.push(s.charCodeAt(i));
            }
            return this;
        };
        Chunk.prototype.writeUInt32At = function (n, pos) {
            if (pos === void 0)
                pos = this.data.length;
            while (this.data.length < pos + 4)
                this.data.push(0);
            this.data[pos + 0] = n >>> 24;
            this.data[pos + 1] = (n >>> 16) & 0xff;
            this.data[pos + 2] = (n >>> 8) & 0xff;
            this.data[pos + 3] = (n) & 0xff;
            return this;
        };
        Chunk.prototype.placeHolder32 = function () {
            var pos = this.data.length;
            this.data.push(0);
            this.data.push(0);
            this.data.push(0);
            this.data.push(0);
            return pos;
        };
        Chunk.prototype.writeVarInt = function (n) {
            var bytes = [];
            bytes.push(n & 0x7f);
            n >>>= 7;
            while (n > 0) {
                bytes.push((n & 0x7f) | 0x80);
                n >>>= 7;
            }
            for (var i = bytes.length - 1; i >= 0; i--) {
                this.data.push(bytes[i]);
            }
            return this;
        };
        Chunk.prototype.pos = function () {
            return this.data.length;
        };
        Chunk.prototype.append = function (c) {
            for (var i = 0; i < c.data.length; i++) {
                this.data.push(c.data[i]);
            }
            return this;
        };
        Chunk.prototype.beginChunk = function (type) {
            this.writeString(type);
            this._lastPos = this.placeHolder32();
        };
        Chunk.prototype.endChunk = function () {
            this.writeUInt32At(this.data.length - this._lastPos - 4, this._lastPos);
        };
        Chunk.prototype.writeTempoEvent = function (tempo) {
            this.writeBytes([0xFF, 0x51, 0x03, (tempo >>> 16) & 0xff, (tempo >>> 8) & 0xff, (tempo) & 0xff]);
        };
        Chunk.prototype.writeKeySignatureEvent = function (shift, minor) {
            this.writeBytes([0xFF, 0x59, 0x02, Note.shiftToKeySignature(shift, minor), minor ? 1 : 0]);
        };
        Chunk.prototype.writeTimeSignatureEvent = function (sig, metronome) {
            this.writeBytes([0xFF, 0x58, 0x04, sig.numerator, log2(sig.denominator), metronome.clocks, metronome.n32]);
        };
        Chunk.prototype.writeEndOfTrack = function () {
            this.writeBytes([0x1, 0xFF, 0x2F, 0x00]);
        };
        return Chunk;
    }());
    function writeHeaderChunk(file, chunk) {
        chunk.writeString('MThd');
        chunk.writeUInt32At(6);
        chunk.writeWord(1);
        chunk.writeWord(file.tracks.length + 1);
        chunk.writeWord(file.division & 0x7fff);
    }
    function log2(n) {
        var i = 0;
        while (n >= 2) {
            n >>= 1;
            i++;
        }
        return i;
    }
    function writeMetaEvents(file, chunk) {
        chunk.writeVarInt(0).writeTempoEvent(file.startTempo);
        chunk.writeVarInt(0).writeKeySignatureEvent(file.keysig, file.minor);
        chunk.writeVarInt(0).writeTimeSignatureEvent(file.timesig, file.metronome);
        chunk.writeEndOfTrack();
    }
    function writeTrackMetaEvents(file, track, channel, chunk) {
        if (track.instrument !== -1)
            chunk.writeBytes([0x00, 0xC0 | channel, track.instrument & 0x7f]);
        chunk.writeBytes([0x00, 0xB0 | channel, 0x79, 0x00]);
        chunk.writeBytes([0x00, 0xB0 | channel, 0x07, track.volume & 0xff]);
        if (track.name !== null) {
            chunk.writeBytes([0x00, 0xFF, 0x03, track.name.length & 0xff]).writeString(track.name, 0xff);
        }
    }
    function writeTrackEvents(file, track, chunk) {
        for (var _i = 0, _b = track.events; _i < _b.length; _i++) {
            var event_1 = _b[_i];
            chunk.writeVarInt(event_1.delta);
            switch (event_1.type) {
                case MidiEventType.NOTEON:
                    chunk.writeBytes([0x90 | event_1.channel, event_1.note, event_1.velocity]);
                    break;
                case MidiEventType.NOTEOFF:
                    chunk.writeBytes([0x80 | event_1.channel, event_1.note, 0]);
                    break;
                case MidiEventType.TEMPO_CHANGE:
                    chunk.writeTempoEvent(event_1.tempo);
                    break;
                case MidiEventType.KEY_SIGNATURE_CHANGE:
                    chunk.writeKeySignatureEvent(event_1.shift, event_1.minor);
                    break;
                case MidiEventType.TIME_SIGNATURE_CHANGE:
                    chunk.writeTimeSignatureEvent(event_1.sig, file.metronome);
                    break;
                default:
                    throw new Error("Unimplemented event " + MidiEventType[event_1.type]);
            }
        }
        chunk.writeBytes([0x1, 0xFF, 0x2F, 0x00]);
    }
    function saveFormat1MidiFile(file) {
        var ret = new Chunk();
        writeHeaderChunk(file, ret);
        ret.beginChunk('MTrk');
        writeMetaEvents(file, ret);
        ret.endChunk();
        for (var i = 0, _a = file.tracks; i < _a.length; i++) {
            ret.beginChunk('MTrk');
            writeTrackMetaEvents(file, _a[i], i, ret);
            writeTrackEvents(file, _a[i], ret);
            ret.endChunk();
        }
        return ret.data;
    }
    function saveMidiFile(file) {
        switch (file.format) {
            case 1: return saveFormat1MidiFile(file);
            default: throw new Error("Format " + file.format + " is not currently supported");
        }
    }



    var debug = /*#__PURE__*/Object.freeze({
        MacroExpander: MacroExpander,
        Scanner: Scanner,
        ErrorReporter: ErrorReporter,
        createParser: createParser,
        dumpNoteList: dumpNoteList,
        createNoteQueue: createNoteQueue,
        eventToString: eventToString,
        Note: Note,
        saveMidiFile: saveMidiFile
    });

    function createContext() {
        var eReporter = new ErrorReporter();
        var scanner = new Scanner();
        var parser = createParser(eReporter);
        var warns = [];
        scanner.macroParamChar = '$';
        return {
            parse: parse,
            getErrors: getErrors,
            hasError: function () { return eReporter.msgs.length > 0; },
            getPrintedErrors: getPrintedErrors
        };
        function parse(input) {
            if (typeof input === 'string') {
                var i_1 = 0;
                scanner.reset({
                    peek: function () { return i_1 >= input.length ? null : input.charAt(i_1); },
                    next: function () { return i_1 >= input.length ? null : input.charAt(i_1++); },
                });
            }
            else {
                scanner.reset(input);
            }
            eReporter.reset();
            warns.length = 0;
            return parser.parse(scanner);
        }
        function getErrors() {
            return eReporter.msgs;
        }
        function getPrintedErrors(lines, marker, space) {
            if (marker === void 0) { marker = '^'; }
            if (space === void 0) { space = ' '; }
            var ret = [];
            for (var _i = 0, _a = eReporter.msgs; _i < _a.length; _i++) {
                var msg = _a[_i];
                ret.push("Error: " + msg.msg + " (line " + msg.range.start.line + ")");
                for (var _b = 0, _c = markLines(lines, msg.range, marker, space); _b < _c.length; _b++) {
                    var line = _c[_b];
                    ret.push(line);
                }
                ret.push('');
            }
            return ret;
        }
    }

    exports.debug = debug;
    exports.createContext = createContext;
    exports.saveMidiFile = saveMidiFile;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=quick-midi.js.map
